/* Duplicate Tab Extension, by Twanno.
 *
 * The duplicateInTab, duplicateInWindow, waitForSessionHistory, copyTabHistory and cloneTabHistory
 * functions have been taken from Pike's Clone Window 0.2.2 extension, and have been adjusted to my needs.
 * Pike's extension: http://www.pikey.me.uk/mozilla/#cw
 */

var duplicateTab = {
  version: '0.8',
  prefService: null,
  prefBranch: null,
  stringBundle: null,
  tablibUsed: false,
  tablibPref: null,
  doSetScrollPosition: true,
  loadURL: null,

  handleLinkClick: null,
  gotoHistoryIndex: null,
  BrowserForward: null,
  BrowserBack: null,

  Consts: {
    TABLIB_PREFS: ['tablib.tabs_open_relative', 'extensions.tabmix.openTabNext', 'extensions.minit.open_tab', 'minit.open_tabs_relative']
  },

  // Setup object properties
  init: function()
  {
    // load preference service
    duplicateTab.prefService = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch(null);
    duplicateTab.prefBranch = duplicateTab.prefService.QueryInterface(Components.interfaces.nsIPrefBranchInternal);
    duplicateTab.prefBranch.addObserver("duplicatetab.contextMenuItems.position", dtContextMenu.positionListener, false);

    // load stringbundle
    duplicateTab.stringBundle = document.getElementById("duplicatetab-string-bundle");

    // check whether Tab Clicking Options is installed
    dtMergeWindows.checkTabClicking();

    // add clicking event listener to tabbar to notify selecting of a tab for merging.
    getBrowser().mStrip.childNodes[2].addEventListener("click", dtMergeWindows.tabBarClick, false);

    // add context menu items to the tab container element
    dtContextMenu.addTabContextMenu();

    //load tablib
    if (window.tablib != undefined) {
      tablib.init();
      duplicateTab.tablibUsed = true;
      for (var i = 0; i < duplicateTab.Consts.TABLIB_PREFS.length; i++) {
        if (gBrowser.addTab.toString().indexOf(duplicateTab.Consts.TABLIB_PREFS[i]) != -1)
          duplicateTab.tablibPref = duplicateTab.Consts.TABLIB_PREFS[i];
      }
    }
    var os = Components.classes["@mozilla.org/observer-service;1"]
                       .getService(Components.interfaces.nsIObserverService);
    os.addObserver(duplicateTab.newWindowObserver, "duplicatetab:session-history-available", false);
    os.addObserver(duplicateTab.duplicateTabObserver, "duplicatetab:duplicate-in-tab", false);
    os.addObserver(duplicateTab.duplicateTabObserver, "duplicatetab:duplicate-in-window", false);
    os.addObserver(duplicateTab.duplicateTabObserver, "duplicatetab:merge-windows", false);

    if (gBrowser.onDragStart && !window.TMonDrop) {
      gBrowser.mStrip.addEventListener("dragdrop", function (event) {nsDragAndDrop.drop(event, duplicateTab.tabBarDNDObserver)}, true);
      gBrowser.mStrip.addEventListener("dragover", function (event) {nsDragAndDrop.dragOver(event, duplicateTab.tabBarDNDObserver)}, true);
      gBrowser.mStrip.addEventListener("dragexit", function (event) {nsDragAndDrop.dragExit(event, duplicateTab.tabBarDNDObserver)}, true);
    }

    // add support for middle clicking on a link to open the link in a duplicated tab
    duplicateTab.handleLinkClick = window.handleLinkClick;
    window.handleLinkClick = duplicateTab.newHandleLinkClick;

    // add support for middle clicking back or forward history to duplicate tab
    duplicateTab.gotoHistoryIndex = window.gotoHistoryIndex;
    window.gotoHistoryIndex = duplicateTab.newGotoHistoryIndex;

    duplicateTab.BrowserForward = window.BrowserForward;
    window.BrowserForward = duplicateTab.newBrowserForward;

    duplicateTab.BrowserBack = window.BrowserBack;
    window.BrowserBack = duplicateTab.newBrowserBack;

    // add support for Mozilla/SeaMonkey to middle click on the history drop down menu items
    var fwContextMenu = document.getElementById('forwardMenu');
    var backContextMenu = document.getElementById('backMenu');
    var fwButton = document.getElementById('forward-button');
    var backButton = document.getElementById('back-button');
    var fwDropDownMenu = fwButton.getElementsByTagName("menupopup")[0];
    var backDropDownMenu = backButton.getElementsByTagName("menupopup")[0];

    if (!fwDropDownMenu.hasAttribute('onclick'))
    {
      fwDropDownMenu.setAttribute('onclick', 'duplicateTab.checkForMiddleClick(this, event)');
      fwDropDownMenu.setAttribute('oncommand', 'gotoHistoryIndex(event); event.preventBubble();');
      fwContextMenu.setAttribute('onclick', 'duplicateTab.checkForMiddleClick(this, event)');
      fwButton.setAttribute('onclick', 'duplicateTab.checkForMiddleClick(this, event)');
    }
    if (!backDropDownMenu.hasAttribute('onclick'))
    {
      backDropDownMenu.setAttribute('onclick', 'duplicateTab.checkForMiddleClick(this, event)');
      backDropDownMenu.setAttribute('oncommand', 'gotoHistoryIndex(event); event.preventBubble();');
      backContextMenu.setAttribute('onclick', 'duplicateTab.checkForMiddleClick(this, event)');
      backButton.setAttribute('onclick', 'duplicateTab.checkForMiddleClick(this, event)');
    }
  },

  deinit: function() {
    duplicateTab.prefBranch.removeObserver("duplicatetab.contextMenuItems.position", dtContextMenu.positionListener);
    getBrowser().mStrip.childNodes[2].removeEventListener("click", dtMergeWindows.tabBarClick, false);
    getBrowser().mStrip.childNodes[2].removeEventListener("click", dtContextMenu.tabBarClick, false);
    dtContextMenu.deinit();
    var os = Components.classes["@mozilla.org/observer-service;1"]
                       .getService(Components.interfaces.nsIObserverService);
    os.removeObserver(duplicateTab.newWindowObserver, "duplicatetab:session-history-available");
    os.removeObserver(duplicateTab.duplicateTabObserver, "duplicatetab:duplicate-in-tab");
    os.removeObserver(duplicateTab.duplicateTabObserver, "duplicatetab:duplicate-in-window");
    os.removeObserver(duplicateTab.duplicateTabObserver, "duplicatetab:merge-windows");

    // restore original functions
    window.handleLinkClick = duplicateTab.handleLinkClick;
    window.gotoHistoryIndex = duplicateTab.gotoHistoryIndex;
    window.BrowserForward = duplicateTab.BrowserForward;
    window.BrowserBack = duplicateTab.BrowserBack;
  },

  newHandleLinkClick: function(event, href, linkNode)
  {
    if (event.button == 1) {
      var openTab;
      try {
        openTab = duplicateTab.prefService.getBoolPref("browser.tabs.opentabfor.middleclick");
      }
      catch(e) {
        openTab = true;
      }

      var duplicateTabForMiddleClick = duplicateTab.prefService.getBoolPref("duplicatetab.link.middleclick.open_in_duplicate_tab");
      var eAccelModifier = ((event.ctrlKey || event.metaKey) && !event.shiftKey && !event.altKey);
      var eAccelShiftModifier = ((event.metaKey || event.ctrlKey) && event.shiftKey && !event.altKey);
      var eNoModifier = (!event.altKey && !event.metaKey && !event.ctrlKey);
      // aEvent.shiftKey is allowed, so the duplicating the link in a new tab follows the focussing
      // of the normal middle click handling (shift key inverses the selection pref for new tabs)

      if (eAccelModifier || (eNoModifier && duplicateTabForMiddleClick && openTab)) {
        // open link in new tab with history
        duplicateTab.openLinkWithHistory('tab', href, event);
        event.preventBubble();
        return true;
      }
      else if (eAccelShiftModifier || (eNoModifier && duplicateTabForMiddleClick && !openTab)) {
        // open link in new window with history
        duplicateTab.openLinkWithHistory('window', href);
        event.preventBubble();
        return true;
      }
    }

    // call original handleLinkClick when the link will not be opened in a duplicate tab
    return duplicateTab.handleLinkClick.apply(this, arguments);
  },

  newGotoHistoryIndex: function(aEvent)
  {
    var index = aEvent.target.getAttribute("index");
    if ('button' in aEvent && aEvent.button == 1 && duplicateTab.openBFInDuplicateTab(index))
      return true;

    // call original gotoHistoryIndex when we are not duplicating the tab
    return duplicateTab.gotoHistoryIndex.apply(this, arguments);
  },

  newBrowserForward: function(aEvent)
  {
    if ('button' in aEvent && aEvent.button == 1 && duplicateTab.openBFInDuplicateTab('forward'))
      return;

    // call original BrowserForward when we are not duplicating the tab
    duplicateTab.BrowserForward.apply(this, arguments);
  },

  newBrowserBack: function(aEvent)
  {
    if ('button' in aEvent && aEvent.button == 1 && duplicateTab.openBFInDuplicateTab('back'))
      return;

    // call original BrowserBack when we are not duplicating the tab
    duplicateTab.BrowserBack.apply(this, arguments);
  },

  openBFInDuplicateTab: function(aIndex)
  {
    var openInDuplicateTab = false;
    try {
      openInDuplicateTab = duplicateTab.prefService.getBoolPref("duplicatetab.bfhistory.middleclick.open_in_duplicate_tab");
    }
    catch(e) {}

    if (openInDuplicateTab) {
      var currentIndex = getBrowser().mCurrentBrowser.webNavigation.sessionHistory.index;
      var currentCount = getBrowser().mCurrentBrowser.webNavigation.sessionHistory.count;

      if (aIndex == 'back')
        aIndex = currentIndex - 1;
      else if (aIndex == 'forward')
        aIndex = currentIndex + 1;

      if (aIndex < 0 || aIndex > currentCount - 1) {
        // this happens when middle clicking on a disabled back or forward
        // toolbar button or menu item
        return false;
      }
      duplicateTab.doSetScrollPosition = false;
      duplicateTab.duplicateInTab(getBrowser().mCurrentTab, false, parseInt(aIndex));
    }
    return openInDuplicateTab;
  },

  checkForMiddleClick: function(node, event)
  {
    if (event.button == 1) {
      var fn = new Function("event", node.getAttribute("oncommand"));
      fn.call(node, event);

      // If the middle-click was on part of a menu, close the menu.
      // (Menus close automatically with left-click but not with middle-click.)
      duplicateTab.closeMenus(event.target);
    }
  },

  closeMenus: function(node)
  {
    if ("tagName" in node) {
      if (node.namespaceURI == "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul"
          && (node.tagName == "menupopup" || node.tagName == "popup"))
        node.hidePopup();

      duplicateTab.closeMenus(node.parentNode);
    }
  },

  openLinkWithHistory: function(target, url)
  {
    // disallow invalid urls to be handled
    if (!url || !url.length || url.indexOf(" ", 0) != -1 || /^\s*javascript:/.test(url))
    {
      // show message that hyperlink contains an invalid url and can't be loaded
      var notificationTopic = "duplicatetab:duplicate-in-" + target;
      var errorimage = "chrome://global/skin/icons/Warning.png";
      var message = duplicateTab.stringBundle.getString('duplicatetab.invalid-url.warning'); // "The hyperlink contains an invalid URL, and can not be opened in a duplicate tab."
      var buttonString = duplicateTab.stringBundle.getString('duplicatetab.invalid-url.solution'); // "Duplicate the tab anyway"
      var buttonAccessKey = 'D'; // TODO localize this accesskey
      if (getBrowser().showMessage) {
        // Firefox 1.0 / 1.5
        getBrowser().showMessage(gBrowser.selectedBrowser, errorimage , message, buttonString,
                                 window, notificationTopic, null,
                                 "top", true);
      }
      else if (getBrowser().getNotificationBox) {
        // Firefox 2.0
        var notificationBox = getBrowser().getNotificationBox();
        var notification = notificationBox.getNotificationWithValue(notificationTopic);
        if (notification)
          notification.label = message;
        else {
          var buttons = [{
            label: buttonString,
            accessKey: buttonAccessKey,
            popup: null,
            callback: function() {
              var os = Components.classes["@mozilla.org/observer-service;1"]
                       .getService(Components.interfaces.nsIObserverService);
              os.notifyObservers(window, notificationTopic, "");
            }
          }];

          const priority = notificationBox.PRIORITY_WARNING_LOW;
          notificationBox.appendNotification(message, notificationTopic, errorimage, priority, buttons);
        }
      }
      return;
    }

    this.loadURL = url

    switch (target) {
      case 'window':
        this.duplicateInWindow(gBrowser.mCurrentTab);
        break;
      case 'tab':
        var oldTab = gBrowser.mCurrentTab;
        var newTab = gBrowser.addTab();

        if (gBrowser.moveTabTo && this.prefService.getBoolPref("duplicatetab.tablib.enabled")) {
          var oldTabIndex = (gBrowser.getTabIndex)? gBrowser.getTabIndex(oldTab): oldTab._tPos;
          gBrowser.moveTabTo((gBrowser.getTabIndex)? gBrowser.getTabIndex(newTab): newTab, oldTabIndex + 1);
        }

        // follow preferences for opening a link in a duplicate tab when arguments[2] is specified
        if (arguments[2])
        {
          var loadInBackGround = true;
          try {
            loadInBackGround = this.prefService.getBoolPref("browser.tabs.loadInBackground")
          }
          catch(e) {}
          // reverse focussing of the new tab when the shiftKey was pressed
          if (arguments[2] instanceof Event && 'shiftKey' in arguments[2] && arguments[2].shiftKey)
            loadInBackGround = !loadInBackGround;
          if (!loadInBackGround)
            gBrowser.selectedTab = newTab;
        }
        else
          gBrowser.selectedTab = newTab;

        var clonedContents = this.getClonedContent(gBrowser.getBrowserForTab(oldTab), true);
        var openedBrowser = gBrowser.getBrowserForTab(newTab);

        clonedContents[0].push(this.createNewSHEntry(this.loadURL));
        clonedContents[0].index++;

        this.setClonedContent(openedBrowser, clonedContents);

        this.loadURL = null;
        break;
      default:
        this.loadURL = null;
        break;
    }
  },

  // Check for duplication to tab or to window in one function where the browser to be duplicated is asigned to a value
  // Argument1 aTarget: 'tab' or 'window', to which duplication must occur
  // Argument2 aTab: optional, tab of the browser to be duplicated. If aTab is ommited, the current tab is used
  duplicateTab: function(aTarget, aTab, aEvent)
  {
    if (!aTab && aEvent && !(aEvent.shiftKey && (aEvent.ctrlKey || aEvent.metaKey))) {
      if (getBrowser().mContextTab && getBrowser().mContextTab.localName == 'tab')
        aTab = getBrowser().mContextTab;
      else if (document.popupNode && document.popupNode.localName == 'tab')
        aTab = document.popupNode;
    }
    if (!aTab)
      aTab = getBrowser().mCurrentTab;


    if (aTarget == 'tab') {
      this.duplicateInTab(aTab);
    }
    else if (aTarget == 'window') {
      this.duplicateInWindow(aTab);
    }
  },

  duplicateInTab: function(aTab, aNewPosition, aNewHistoryIndex)
  {
    // put the contents of the browser to be duplicated into an array
    var clonedContents = this.getClonedContent(getBrowser().getBrowserForTab(aTab));

    // Open new tab
    var newTab = getBrowser().addTab("about:blank");

    // give the new tab focus if prefered by the user
    if (duplicateTab.prefService.getBoolPref("duplicatetab.newtab.focus"))
      getBrowser().selectedTab = newTab;

    // place the duplicate tab at the position passed as an argument to the function
    if (gBrowser.moveTabTo && typeof aNewPosition == "number")
      gBrowser.moveTabTo((gBrowser.getTabIndex)? gBrowser.getTabIndex(newTab): newTab, aNewPosition);

    // place duplicate tab next to original if desired.
    else if (gBrowser.moveTabTo)
    {
      var tablibUsed = (duplicateTab.tablibPref)?duplicateTab.prefService.getBoolPref(duplicateTab.tablibPref):false;
      if (tablibUsed || duplicateTab.prefService.getBoolPref("duplicatetab.tablib.enabled"))
      {
        var newTabPos = (gBrowser.getTabIndex)? gBrowser.getTabIndex(newTab): newTab._tPos;
        var oldTabPos = (gBrowser.getTabIndex)? gBrowser.getTabIndex(aTab): aTab._tPos;
        if (newTabPos > oldTabPos)
          gBrowser.moveTabTo((gBrowser.getTabIndex)? gBrowser.getTabIndex(newTab): newTab, oldTabPos + 1);
        if (newTabPos < oldTabPos)
          gBrowser.moveTabTo((gBrowser.getTabIndex)? gBrowser.getTabIndex(newTab): newTab, oldTabPos);
      }
    }

    var openedBrowser = getBrowser().getBrowserForTab(newTab);

    // for when opening back or forward history in duplicate tab: go to the correct index:
    if (typeof aNewHistoryIndex == "number")
      clonedContents[0].index = aNewHistoryIndex;

    // put the cloned contents into the new browser
    this.setClonedContent(openedBrowser, clonedContents);

    // Select URL bar
    if (gURLBar)
      window.setTimeout('gURLBar.select();', 100);
  },

  duplicateInWindow: function(aTab)
  {
    // get the cloned contents from the tab
    this.newWindowObserver.clonedContents = this.getClonedContent(gBrowser.getBrowserForTab(aTab), Boolean(this.loadURL));


    // Open new window
    var chromeURL = duplicateTab.prefService.getCharPref("browser.chromeURL");
    this.newWindowObserver.openedWindow = window.openDialog(chromeURL, '_blank', 'chrome,all,dialog=no');

    // Wait until session history is created in new window
    this.newWindowObserver.openedWindow.addEventListener('load', function() {duplicateTab.waitForSessionHistory(10);}, false);
  },

  waitForSessionHistory: function(attempts)
  {
    // Test if sessionHistory exists yet
    var webNav = duplicateTab.newWindowObserver.openedWindow.document.getElementById('content').webNavigation;
    try    {
      webNav.sessionHistory;
    }

    // webNav.sessionHistory is not yet available, try again later
    catch (err) {
      if (attempts)
        window.setTimeout(duplicateTab.waitForSessionHistory, 100, --attempts);
      return;
    }
    if ((webNav.sessionHistory == null) && attempts) {
      window.setTimeout(duplicateTab.waitForSessionHistory, 100, --attempts);
      return;
    }

    // notify the observers that session history is available in openedBrowser, so the cloned content can be added
    var os = Components.classes["@mozilla.org/observer-service;1"]
                       .getService(Components.interfaces.nsIObserverService);
    os.notifyObservers(null, "duplicatetab:session-history-available", "");
  },

  /*
   * Create a new nsISHEntry from an href
   *    @argument href: a string representing the URI for which the history entry should be created
   */
  createNewSHEntry: function(href) {
    var newSHEntry = Components.classes["@mozilla.org/browser/session-history-entry;1"]
                               .createInstance(Components.interfaces.nsISHEntry)
                               .QueryInterface(Components.interfaces.nsIHistoryEntry);
    var ioService = Components.classes["@mozilla.org/network/io-service;1"]
                              .getService(Components.interfaces.nsIIOService);

    newSHEntry.setURI(ioService.newURI(href, null, null));
    newSHEntry.setTitle(href);
    newSHEntry.setIsSubFrame(false);
    newSHEntry.loadType = Components.interfaces.nsIDocShellLoadInfo.loadHistory;

    return newSHEntry;
  },

  getHistoryRange: function(aIndex, aCount, aOnlyBack)
  {
    var backEnabled = duplicateTab.prefService.getBoolPref("duplicatetab.back-history.enabled");
    var forwardEnabled = duplicateTab.prefService.getBoolPref("duplicatetab.forward-history.enabled");
    var forwardLength = duplicateTab.prefService.getIntPref("duplicatetab.forward-history.length");
    var backLength = duplicateTab.prefService.getIntPref("duplicatetab.back-history.length");

    var currentIndex = aIndex;
    var lastIndex = aCount - 1;
    var firstIndex;
    var finalLength;

    if (backEnabled) {
      if (backLength == 0 || backLength >= currentIndex)
        firstIndex = 0;
      else
        firstIndex = currentIndex - backLength;
    }
    else
      firstIndex = currentIndex;

    if (forwardEnabled && !aOnlyBack) {
      if (forwardLength == 0 || currentIndex + forwardLength >= lastIndex)
        finalLength = lastIndex + 1;
      else
        finalLength = currentIndex + forwardLength + 1;
    }
    else
      finalLength = currentIndex + 1;

    var finalIndex = currentIndex - firstIndex;

    return {start: firstIndex, index: finalIndex, length: finalLength};
  },

  // copy a sessionHistory and put it into an array
  // Argument1 originalHistory: webNavigation.sessionHistory browser history to be copied
  // returns: an array containing a copy of the history
  copyTabHistory: function(originalHistory, aOnlyBack)
  {
    var range = this.getHistoryRange(originalHistory.index, originalHistory.count, aOnlyBack);

    var copiedHistory = new Array();
    for (var i = range.start; i < range.length; i++) {
      copiedHistory.push(originalHistory.getEntryAtIndex(i, false));
    }
    copiedHistory.index = range.index;

    return copiedHistory;
  },

  cloneHistoryEntry: function(aEntry) {
    if (!aEntry)
      return null;
    aEntry = aEntry.QueryInterface(Components.interfaces.nsISHContainer);
    var newEntry = aEntry.clone();
    newEntry = newEntry.QueryInterface(Components.interfaces.nsISHContainer);
    newEntry.loadType = Math.floor(aEntry.loadType);
    if (aEntry.childCount) {
      for (var j = 0; j < aEntry.childCount; j++) {
          var childEntry = duplicateTab.cloneHistoryEntry(aEntry.GetChildAt(j));
          if (childEntry)
            newEntry.AddChild(childEntry, j);
      }
    }
    return newEntry;
  },

  // Clone an array of history entries into a browsers webNavigation.sessionHistory
  // Argument1 webNav: The webNavigation object of a newly created browser.
  // Argument2 originalHistory: an array containing history entries from the original browser
  cloneTabHistory: function(webNav, originalHistory)
  {
    var newHistory = webNav.sessionHistory

    newHistory.QueryInterface(Components.interfaces.nsISHistoryInternal);

    // delete history entries if they are present
    if (newHistory.count > 0) newHistory.PurgeHistory(newHistory.count);

    for (var i = 0; i < originalHistory.length; i++) {
      var entry = originalHistory[i].QueryInterface(Components.interfaces.nsISHEntry);
      var newEntry = duplicateTab.cloneHistoryEntry(entry);
      if (newEntry)
        newHistory.addEntry(newEntry, true);
    }

    // Go to current history location
    if (originalHistory.index < originalHistory.length)
      gotoHistoryIndex(10);

    function gotoHistoryIndex(attempts) {
      try {
        webNav.gotoIndex(originalHistory.index);
      }
      catch(e) {
        // do some math to increase the timeout each time we try to update the history index
        if (attempts)
          setTimeout(gotoHistoryIndex, (11 - attempts) * (15 - attempts), --attempts);
      }
    }
  },

  // aBrowser: xul:browser which should be cloned.
  // aOnlyBack: true if forward history should not be cloned.
  getClonedContent: function(aBrowser, aOnlyBack)
  {
    var clonedContents = new Array();
    clonedContents[0] = this.copyTabHistory(aBrowser.webNavigation.sessionHistory, aOnlyBack);
    clonedContents[1] = aBrowser.contentWindow.scrollX;
    clonedContents[2] = aBrowser.contentWindow.scrollY;
    clonedContents[3] = aBrowser.markupDocumentViewer.textZoom;

    if (aBrowser.thumbcache)
      clonedContents[4] = this.cloneThumbcache(aBrowser, aOnlyBack);

    return clonedContents;
  },

  // support for Reveal extension: clone thumbnails
  cloneThumbcache: function(aBrowser, aOnlyBack)
  {
    var i, thumbnails = new Array();
    var range = this.getHistoryRange(aBrowser.thumbcache.index, aBrowser.thumbcache.thumbnails.length, aOnlyBack);

    for (i = range.start; i < range.length; i++) {
      try {
        var thumbnail = aBrowser.thumbcache.thumbnails[i];
        if (thumbnail.thumb) {
          if (thumbnail.thumb.height == 1) {
            var oldIndex = aBrowser.thumbcache.index;
            aBrowser.thumbcache.index = i;
            aBrowser.thumbcache.makeThumbnail();
            aBrowser.thumbcache.index = oldIndex;
          }
          var clonedThumb = thumbnail.thumb.cloneNode(true);
          // canvas context is not cloned, clone it now:
          var ctx = clonedThumb.getContext("2d");
          ctx.drawImage(thumbnail.thumb, 0, 0, clonedThumb.width, clonedThumb.height);
          var clonedTitle = thumbnail.title;
          var clonedURI = thumbnail.uri;
          var clonedHost = thumbnail.host;
          var clonedPagetext = thumbnail.pagetext;
          var clonedLabelValue = thumbnail.label.getAttribute("value");
          var clonedLabelTooltipText = thumbnail.label.getAttribute("tooltiptext");
          var clonedIconsource = thumbnail.icon.getAttribute("src");
          thumbnails.push([clonedThumb, clonedTitle, clonedURI, clonedHost, clonedPagetext,
                           clonedLabelValue, clonedLabelTooltipText, clonedIconsource])
        }
      }
      catch(e) {
        // just fail.
      }
    }
    return thumbnails;
  },

  // sets previously cloned content into a new browser
  // Argument1 aBrowser: a newly created xul:browser to set the contents to
  // Argument2 aClonedContents: an Array containing browser history, scrollposition and zoomfactor, to be set to the browser
  setClonedContent: function(aBrowser, aClonedContents)
  {
    function setTextZoom(attempts) {
      aBrowser.markupDocumentViewer.textZoom = aClonedContents[3];
      if (attempts && aBrowser.markupDocumentViewer.textZoom != aClonedContents[3])
        setTimeout(setTextZoom, 10, --attempts)
    }

    function setScrollPosition(attempts) {
      aBrowser.contentWindow.scrollTo(aClonedContents[1], aClonedContents[2]);
      if (attempts && (aBrowser.contentWindow.scrollX != aClonedContents[1] || aBrowser.contentWindow.scrollY != aClonedContents[2]))
        setTimeout(setScrollPosition, 10, --attempts);
    }

    if (aClonedContents[0].length == 0)
      return;

    this.cloneTabHistory(aBrowser.webNavigation, aClonedContents[0]);

    // for reveal extension: clone thumbnails for history entries
    if (aClonedContents[4])
      this.handleThumbCache(aClonedContents[4], aBrowser.thumbcache);

    if (this.prefService.getBoolPref("duplicatetab.scrollposition.enabled") && this.doSetScrollPosition)
      setScrollPosition(15);

    if (this.prefService.getBoolPref("duplicatetab.zoomfactor.enabled"))
      setTextZoom(15);

    if (!this.doSetScrollPosition)
      this.doSetScrollPosition = true;
  },

  // duplicate thumbnails for reveal extension (http://forums.mozillazine.org/viewtopic.php?p=2046293).
  handleThumbCache: function(thumbnails, newThumbcache)
  {
    for (var i = 0; i < thumbnails.length; i++) {
      try {
        cloneThumbNail(thumbnails[i], newThumbcache.thumbnails[i]);
      }
      catch(e) {
        // just fail.
      }
    }

    function cloneThumbNail(oldThumbnail, newThumbnail) {
      newThumbnail.title = oldThumbnail[1];
      newThumbnail.uri = oldThumbnail[2];
      newThumbnail.host = oldThumbnail[3];
      newThumbnail.pagetext = oldThumbnail[4];
      newThumbnail.label.setAttribute("value", oldThumbnail[5]);
      newThumbnail.label.setAttribute("tooltiptext", oldThumbnail[6]);
      newThumbnail.icon.setAttribute("src", oldThumbnail[7]);
      var newThumb = newThumbnail.thumb;
      var thumbClone = oldThumbnail[0];
      var thumbCloneStyle = thumbClone.getAttribute("style");
      newThumb.setAttribute("style", thumbCloneStyle);
      newThumb.width = thumbClone.width;
      newThumb.height = thumbClone.height;
      var ctx = newThumb.getContext("2d");
      ctx.drawImage(thumbClone, 0, 0, newThumb.width, newThumb.height);
    }
  },

  checkLinkNode: function(target)
  {
    var linkNode;

    if (target instanceof HTMLAnchorElement ||
        target instanceof HTMLAreaElement ||
        target instanceof HTMLLinkElement) {
          if (target.hasAttribute("href"))
            linkNode = target;
    }
    else {
      linkNode = target;
      while (linkNode && !(linkNode instanceof HTMLAnchorElement))
        linkNode = linkNode.parentNode;
      if (linkNode && !linkNode.hasAttribute("href"))
        linkNode = null;
    }

    return linkNode;
  },

  tabButtonObserver:
  {
    onDragOver: function(aEvent, aFlavour, aDragSession)
    {
      aEvent.target.setAttribute("dragover", "true");
      return true;
    },
    onDragExit: function (aEvent, aDragSession)
    {
      aEvent.target.removeAttribute("dragover");
    },
    onDrop: function (aEvent, aXferData, aDragSession)
    {
      var xferFlavourType = aXferData.flavour.contentType;
      if (xferFlavourType == 'text/x-moz-tab' || xferFlavourType == 'text/x-minit-tab')
      {
        var tab = getBrowser().mTabs[aXferData.data];
        duplicateTab.duplicateTab('tab', tab, aEvent);
      }
      else
      {
        var xferData = aXferData.data.split("\n");
        var url = xferData[0] ? xferData[0] : xferData[1];

        if (aDragSession.sourceNode && aDragSession.sourceNode.parentNode == getBrowser().mTabContainer) {
          duplicateTab.duplicateTab('tab', aDragSession.sourceNode, aEvent);
        }
        else if (url == gBrowser.getBrowserForTab(gBrowser.mCurrentTab).currentURI.spec) {
          duplicateTab.duplicateTab('tab', getBrowser().mCurrentTab, aEvent);
        }
        else if (url) {
          var linkNode;
          if (aDragSession.sourceNode)
            linkNode = duplicateTab.checkLinkNode(aDragSession.sourceNode)
          if (linkNode && window.markLinkVisited)
            markLinkVisited(url, linkNode);
          duplicateTab.openLinkWithHistory('tab', url);
        }
      }
    },
    getSupportedFlavours: function ()
    {
      var flavourSet = new FlavourSet();
      flavourSet.appendFlavour("text/x-moz-tab");
      flavourSet.appendFlavour("text/x-minit-tab");
      flavourSet.appendFlavour("text/unicode");
      flavourSet.appendFlavour("text/html");
      return flavourSet;
    }
  },

  windowButtonObserver: {
    onDragOver: function(aEvent, aFlavour, aDragSession)
    {
      aEvent.target.setAttribute("dragover", "true");
      return true;
    },
    onDragExit: function (aEvent, aDragSession)
    {
      aEvent.target.removeAttribute("dragover");
    },
    onDrop: function (aEvent, aXferData, aDragSession)
    {
      var xferFlavourType = aXferData.flavour.contentType;
      if (xferFlavourType == 'text/x-moz-tab' || xferFlavourType == 'text/x-minit-tab')
      {
        var tab = getBrowser().mTabs[aXferData.data];
        duplicateTab.duplicateTab('window', tab, aEvent);
      }
      else {
        var xferData = aXferData.data.split("\n");
        var url = xferData[0] ? xferData[0] : xferData[1];
        if (aDragSession.sourceNode && aDragSession.sourceNode.parentNode == gBrowser.mTabContainer) {
          duplicateTab.duplicateTab('window', aDragSession.sourceNode, aEvent);
        }
        else if (url == gBrowser.getBrowserForTab(gBrowser.mCurrentTab).currentURI.spec) {
          duplicateTab.duplicateTab('window', gBrowser.mCurrentTab, aEvent);
        }
        else if (url) {
          var linkNode;
          if (aDragSession.sourceNode)
            linkNode = duplicateTab.checkLinkNode(aDragSession.sourceNode)
          if (linkNode && window.markLinkVisited)
            markLinkVisited(url, linkNode);
          duplicateTab.openLinkWithHistory('window', url);
        }
      }
    },
    getSupportedFlavours: function ()
    {
      var flavourSet = new FlavourSet();
      flavourSet.appendFlavour("text/x-moz-tab");
      flavourSet.appendFlavour("text/x-minit-tab");
      flavourSet.appendFlavour("text/unicode");
      flavourSet.appendFlavour("text/html");
      return flavourSet;
    }
  },

  newWindowObserver:
  {
    clonedContents: null,
    openedWindow: null,

    observe: function(aSubject, aTopic, aData)
    {
      switch (aTopic) {
        case "duplicatetab:session-history-available":
          // only the window with this.openedWindow set will
          // respond to this notification.
          if (this.openedWindow) {
            openedBrowser = this.openedWindow.getBrowser().mCurrentBrowser;

            if (duplicateTab.loadURL) {
              this.clonedContents[0].push(duplicateTab.createNewSHEntry(duplicateTab.loadURL));
              this.clonedContents[0].index++;
              duplicateTab.loadURL = null;
            }

            duplicateTab.setClonedContent(openedBrowser, this.clonedContents);

            this.openedWindow = null;
            this.clonedContents = null;
          }
          break;
      }
    }
  },

  duplicateTabObserver: {
    observe: function(aSubject, aTopic, aData) {
      // aSubject should be a window object
      switch (aTopic)
      {
        case "duplicatetab:duplicate-in-tab":
          if (window == aSubject) {
            if (getBrowser().hideMessage)
              getBrowser().hideMessage(getBrowser().mCurrentBrowser, 'top');
            duplicateTab.duplicateTab('tab', null, null);
          }
          break;
        case "duplicatetab:duplicate-in-window":
          if (window == aSubject) {
            if (getBrowser().hideMessage)
              getBrowser().hideMessage(getBrowser().mCurrentBrowser, 'top');
            duplicateTab.duplicateTab('window', null, null);
          }
          break;
        case "duplicatetab:merge-windows":
          if (window == aSubject)
            dtMergeWindows.mergeWindows();
          break;
      }
    }
  },

  tabBarDNDObserver: {
    onDragOver: function (aEvent, aXferData, aDragSession)
    {
      if (!gBrowser.mTabDropIndicatorBar)
        return nsDragAndDrop.dragOver(aEvent, gBrowser);
      if ('shiftKey' in aEvent && aEvent.shiftKey)
        gBrowser.mTabDropIndicatorBar.firstChild.setAttribute('duplicatetab', 'true');
      else
        gBrowser.mTabDropIndicatorBar.firstChild.removeAttribute('duplicatetab');
      return nsDragAndDrop.dragOver(aEvent, gBrowser);
    },
    onDrop: function (aEvent, aXferData, aDragSession)
    {
      if ('shiftKey' in aEvent && aEvent.shiftKey
          && aDragSession.sourceNode && aDragSession.sourceNode.parentNode == getBrowser().mTabContainer)
      {
        var newIndex = (gBrowser.getDropIndex)? gBrowser.getDropIndex(aEvent): gBrowser.getNewIndex(aEvent);
        duplicateTab.duplicateInTab(aDragSession.sourceNode, newIndex);
        return true;
      }
      else
        return nsDragAndDrop.drop(aEvent, gBrowser);
    },
    onDragExit: function (aEvent, aDragSession)
    {
      if (gBrowser.mTabDropIndicatorBar)
        gBrowser.mTabDropIndicatorBar.firstChild.removeAttribute('duplicatetab');
      return nsDragAndDrop.dragExit(aEvent, gBrowser);
    },
    getSupportedFlavours: function ()
    {
      var flavourSet = new FlavourSet();
      flavourSet.appendFlavour("text/unicode");
      flavourSet.appendFlavour("text/html");
      return flavourSet;
    }
  },

  selectBrowser: function() {
    // maintain support for older (< 0.6.5) versions of Tab Clicking Options
    dtMergeWindows.selectBrowser(this.selectedTab);
    delete this.selectedTab
  }
}

window.addEventListener("load", duplicateTab.init, false);
window.addEventListener("unload", duplicateTab.deinit, false);
